/*------------------------------------------------------------------------------
* rinex.c : input and output rinex observation data/navigation messages
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* reference :
*     [1] W.Gurtner and L.Estey, RINEX: The Receiver Independent Exchange Format
*         Version 2.11, 10 December, 2007
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2006/01/16 1.0 new
*           2007/03/14 1.1 read P1 if no obstype of C1
*           2007/04/27 1.2 add readrnxt() function
*           2007/05/25 1.3 add support of file path with wild-card (*)
*                          add support of compressed files
*           2007/11/02 1.4 support sbas/geo satellite
*                          support doppler observables
*                          support rinex bug of week handover
*                          add rinex obs/nav output functions
*           2008/06/16 1.5 export readrnxf(), add compress()
*                          separate sortobs(), uniqeph(), screent()
*           2008/10/28 1.6 fix bug on reading rinex obs header types of observ
*           2009/04/09 1.7 support rinex 2.11
*                          change api of outrnxobsh(),outrnxobsb(),outrnxnavb()
*           2009/06/02 1.8 add api outrnxgnavb()
*           2009/08/15 1.9 support glonass
*                          add slip save/restore functions
*           2010/03/03 1.10 fix bug of array access by disabled satellite
*-----------------------------------------------------------------------------*/
#include <ctype.h>
#include "rtklib.h"

static const char rcsid[]="$Id: rinex.c,v 1.1 2008/07/17 21:48:06 ttaka Exp ttaka $";

#define RNXVER      2.11            /* output rinex format version */

/* read rinex header ---------------------------------------------------------*/
static char readrnxh(FILE *fp, char tobs[][3], int *nt, double *ion,
                     double *utc, int *leaps, double *pos, double *adel)
{
    int i,j,n;
    char buff[256],trnx=' ';
    
    trace(3,"readrnxh:\n");
    
    *nt=0;
    while (fgets(buff,sizeof(buff),fp)) {
        if (strlen(buff)<=60) continue;
        else if (strstr(buff+60,"RINEX VERSION / TYPE")) {
            trnx=*(buff+20);
        }
        else if (strstr(buff+60,"# / TYPES OF OBSERV")) {
            n=(int)str2num(buff,0,6);
            for (i=0,j=10;i<n;i++,j+=6) {
                if (j>58) {if (!fgets(buff,sizeof(buff),fp)) break; j=10;}
                if (i<MAXOBSTYPE) {
                    strncpy(tobs[i],buff+j,2); tobs[i][2]='\0';
                }
            }
            *nt=i;
        }
        else if (strstr(buff+60,"APPROX POSITION XYZ")) {
            for (i=0,j=0;i<3;i++,j+=14) pos[i]=str2num(buff,j,14);
        }
        else if (strstr(buff+60,"ANTENNA: DELTA H/E/N")) {
            for (i=0,j=0;i<3;i++,j+=14) adel[i]=str2num(buff,j,14);
        }
        else if (strstr(buff+60,"ION ALPHA")) {
            for (i=0,j=3;i<4;i++,j+=12) ion[i]=str2num(buff,j,11);
        }
        else if (strstr(buff+60,"ION BETA")) {
            for (i=0,j=3;i<4;i++,j+=12) ion[i+4]=str2num(buff,j,11);
        }
        else if (strstr(buff+60,"DELTA-UTC: A0,A1,T,W")) {
            for (i=0,j=3;i<2;i++,j+=19) utc[i]=str2num(buff,j,19);
            for (;i<4;i++,j+=9) utc[i]=str2num(buff,j,9);
        }
        else if (strstr(buff+60,"LEAP SECONDS")) {
            *leaps=(int)str2num(buff,0,6);
        }
        else if (strstr(buff+60,"END OF HEADER")) break;
    }
    return trnx;
}
/* read rinex observation data -----------------------------------------------*/
static int readrnxo(FILE *fp, gtime_t ts, gtime_t te, double tint,
                    char tobs[][3], int nt, int rcv, obs_t *obs)
{
    const char ftype[]={'1','2','5','7','6',0};
    gtime_t time;
    obsd_t *data;
    double value;
    int i,j,k,n=0,event=0,index[MAXOBSTYPE],sats[MAXOBS];
    char buff[256],satid[4]="";
    unsigned char lock,slips[MAXSAT][NFREQ]={{0}};
    
    trace(3,"readrnxo: nt=%d\n",nt);
    
    for (i=0;i<nt&&i<MAXOBSTYPE;i++) {
        for (j=0;ftype[j];j++) if (tobs[i][1]==ftype[j]) break;
        index[i]=ftype[j]&&j<NFREQ?j:-1;
    }
    i=-1;
    while (fgets(buff,sizeof(buff),fp)) {
        if (i<0) {
            n=(int)str2num(buff,29,3);
            if ((event=(int)str2num(buff,27,2))==4) ; /* header info follows */
            else if (!str2time(buff,0,26,&time)) {
                for (j=0,k=32;j<n;j++,k+=3) {
                    if (k>=68) {if (!fgets(buff,sizeof(buff),fp)) break; k=32;}
                    strncpy(satid,buff+k,3);
                    sats[j]=satid2no(satid);
                }
            }
            else continue;
        }
        else if (event!=4&&sats[i]>=0) {
            if (obs->nmax<=obs->n) {
                if (obs->nmax==0) obs->nmax=65536; else obs->nmax+=65536;
                if (!(obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*obs->nmax))) {
                    trace(1,"readrnxo: memory allocation error (n=%d)\n",obs->nmax);
                    obs->n=obs->nmax=0;
                    return -1;
                }
            }
            data=&obs->data[obs->n];
            data->time=time;
            data->sat=sats[i];
            data->rcv=(unsigned char)rcv;
            for (j=0;j<NFREQ;j++) {
                data->L[j]=data->P[j]=0.0;
                data->D[j]=0.0;
                data->SNR[j]=data->LLI[j]=0;
            }
            for (j=k=0;j<nt;j++,k+=16) {
                if (k>=80) {if (!fgets(buff,sizeof(buff),fp)) break; k=0;}
                value=str2num(buff,k,14);
                if (fabs(value)<1E-6) value=0.0;
                lock=(unsigned char)str2num(buff,k+14,1)&3; /* 1:loss of lock */
                                                            /* 2:halfcycle amb */
                if (j>=MAXOBSTYPE||index[j]<0) continue;
                switch (tobs[j][0]) {
                    case 'L': data->L[index[j]]=value; data->LLI[index[j]]=lock; break;
                    case 'C': data->P[index[j]]=value; break;
                    case 'P': if (data->P[index[j]]==0.0) data->P[index[j]]=value; break;
                    case 'D': data->D[index[j]]=(float)value; break;
                    case 'S': data->SNR[index[j]]=(unsigned char)value; break;
                }
            }
            if (data->sat) {
                /* save slip conditions */
                for (j=0;j<NFREQ;j++) {
                    if (data->LLI[j]&1) slips[data->sat-1][j]|=1;
                }
                if (screent(time,ts,te,tint)) {
                    
                    /* restore slip conditions */
                    for (j=0;j<NFREQ;j++) {
                        if (slips[data->sat-1][j]&1) data->LLI[j]|=1;
                        slips[data->sat-1][j]=0;
                    }
                    obs->n++;
                }
            }
        }
        if (++i>=n) i=-1;
    }
    trace(3,"readrnxo: nobs=%d\n",obs->n);
    return 0;
}
/* adjust time considering week handover -------------------------------------*/
static gtime_t adjweek(gtime_t t, gtime_t t0)
{
    double tt=timediff(t,t0);
    if (tt<-302400.0) return timeadd(t, 604800.0);
    if (tt> 302400.0) return timeadd(t,-604800.0);
    return t;
}
/* read rinex navigation message ---------------------------------------------*/
static int readrnxn(FILE *fp, nav_t *nav, int sys)
{
    gtime_t toc;
    eph_t *eph;
    double e[64];
    int i=0,j,sat,prn=0,line=0;
    char buff[256];
    
    trace(3,"readrnxn: sys=%d\n",sys);
    
    while (fgets(buff,sizeof(buff),fp)) {
        line++;
        if (i==0) {
            if (strlen(buff)<2) {
                trace(1,"rinex nav null line: line=%d\n",line);
                continue;
            }
            prn=(int)str2num(buff,0,2);
            if (str2time(buff,3,17,&toc)) {
                trace(1,"rinex nav time format error: line=%d\n",line);
                continue;
            }
        }
        for (j=i==0?22:3;j<=60;j+=19) e[i++]=str2num(buff,j,19);
        if (i>=31) {
            if (nav->nmax<=nav->n) {
                if (nav->nmax==0) nav->nmax=1024; else nav->nmax+=1024;
                if (!(nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*nav->nmax))) {
                    trace(1,"readrnxn: memory allocation error (n=%d)\n",nav->nmax);
                    nav->n=nav->nmax=0;
                    return -1;
                }
            }
            if ((sat=satno(sys,prn))) {
                eph=&nav->eph[nav->n++];
                eph->sat=sat; eph->iode=(int)e[3]; eph->iodc=(int)e[26];
                eph->sva=(int)e[23]; eph->svh=(int)e[24];
                eph->toc=toc; 
                eph->toe=adjweek(gpst2time((int)e[21],e[11]),toc);
                eph->ttr=adjweek(gpst2time((int)e[21],e[27]),toc);
                eph->A=e[10]*e[10]; eph->e=e[8]; eph->i0  =e[15]; eph->OMG0=e[13];
                eph->omg =e[17]; eph->M0 =e[ 6]; eph->deln=e[ 5]; eph->OMGd=e[18];
                eph->idot=e[19]; eph->crc=e[16]; eph->crs =e[ 4]; eph->cuc =e[ 7];
                eph->cus =e[ 9]; eph->cic=e[12]; eph->cis =e[14]; eph->toes=e[11];
                eph->f0  =e[ 0]; eph->f1 =e[ 1]; eph->f2  =e[ 2]; eph->tgd =e[25];
                eph->fit =e[28];
                eph->l2code=(int)e[20]; eph->l2pflg=(int)e[22];
            }
            else trace(1,"rinex nav sat number error: line=%d sys=%d prn=%d\n",line,sys,prn);
            i=0;
        }
    }
    trace(3,"readrnxn: nnav=%d\n",nav->n);
    return 0;
}
/* read rinex glonass navigation message -------------------------------------*/
static int readrnxg(FILE *fp, nav_t *nav)
{
    gtime_t toe;
    geph_t *geph;
    double e[64],tow,tod;
    int i=0,j,sat,prn=0,week,line=0,day;
    char buff[256];
    
    trace(3,"readrnxg: ngmax=%d\n",nav->ngmax);
    
    while (fgets(buff,sizeof(buff),fp)) {
        line++;
        if (i==0) {
            prn=(int)str2num(buff,0,2);
            if (str2time(buff,3,17,&toe)) { /* toe in utc */
                trace(1,"rinex gnav time format error: line=%d\n",line);
                continue;
            }
            /* round to nearest 15 min interval for abnormal ephemeris */
            tow=time2gpst(toe,&week);
            tow=floor((tow+450.0)/900.0)*900;
            toe=gpst2time(week,tow);
        }
        for (j=i==0?22:3;j<=60;j+=19) e[i++]=str2num(buff,j,19);
        if (i>=15) {
            if (nav->ngmax<=nav->ng) {
                if (nav->ngmax==0) nav->ngmax=1024; else nav->ngmax+=1024;
                if (!(nav->geph=(geph_t *)realloc(nav->geph,sizeof(geph_t)*nav->ngmax))) {
                    trace(1,"readrnxg: memory allocation error (n=%d)\n",nav->ngmax);
                    nav->ng=nav->ngmax=0;
                    return -1;
                }
            }
            if ((sat=satno(SYS_GLO,prn))) {
                geph=&nav->geph[nav->ng++];
                geph->sat=sat;
                geph->toe=utc2gpst(toe); /* utc -> gpst */
                geph->taun=-e[0]; /* -TauN */
                geph->gamn=e[1]; /* +GammaN */
                tow=time2gpst(toe,&week);
                day=(int)floor(tow/86400.0);
                tod=tow-day*86400.0;
                if      (e[2]<tod-43200.0) e[2]+=86400.0;
                else if (e[2]>tod+43200.0) e[2]-=86400.0;
                geph->tof=utc2gpst(gpst2time(week,e[2]+day*86400.0));
                geph->pos[0]=e[ 3]*1E3; /* km -> m */
                geph->vel[0]=e[ 4]*1E3; /* km/s -> m/s */
                geph->acc[0]=e[ 5]*1E3; /* km/s^2 -> m/s^2 */
                geph->pos[1]=e[ 7]*1E3;
                geph->vel[1]=e[ 8]*1E3;
                geph->acc[1]=e[ 9]*1E3;
                geph->pos[2]=e[11]*1E3;
                geph->vel[2]=e[12]*1E3;
                geph->acc[2]=e[13]*1E3;
                geph->svh=(int)e[ 6];
                geph->frq=(int)e[10];
                geph->age=(int)e[14];
                
                /* some receivers do not output minus freq-number */
                if (geph->frq>128) geph->frq-=256;
            }
            else trace(1,"rinex gnav sat number error: line=%d prn=%d\n",line,prn);
            i=0;
        }
    }
    trace(3,"readrnxg: ngnav=%d\n",nav->ng);
    return 0;
}
/* read rinex file -----------------------------------------------------------*/
static char readrnxfp(FILE *fp, gtime_t ts, gtime_t te, double tint, int rcv, 
                      obs_t *obs, nav_t *nav)
{
    double ion[8]={0},utc[4]={0},pos[3]={0},adel[3]={0};
    int i,ntype,sys,leaps=0;
    char type,tobs[MAXOBSTYPE][3];
    
    trace(3,"readrnxfp:\n");
    
    type=readrnxh(fp,tobs,&ntype,ion,utc,&leaps,pos,adel);
    
    if (obs&&0<rcv&&rcv<=MAXRCV&&type=='O') {
        if (norm(pos,3)>0.0) {
            for (i=0;i<3;i++) obs->pos [rcv-1][i]=pos [i];
        }
        if (norm(adel,3)>0.0) {
            for (i=0;i<3;i++) obs->adel[rcv-1][i]=adel[i];
        }
    }
    if (nav&&(type=='N'||type=='L'||type=='Q'||type=='G')) {
        if (norm(ion,8)>0.0) {
            for (i=0;i<8;i++) nav->ion[i]=ion[i];
        }
        if (norm(utc,4)>0.0) {
            for (i=0;i<4;i++) nav->utc[i]=utc[i];
        }
        if (leaps!=0) nav->leaps=leaps;
    }
    switch (type) {
        case 'O': /* observation data */
            if (rcv<=MAXRCV&&obs) {
                if (readrnxo(fp,ts,te,tint,tobs,ntype,rcv,obs)<0) return 'E';
            }
            break;
        case 'N': /* gps navigation */
        case 'L': /* galileo navigation (RINEX 2.11) */
        case 'Q': /* qzs navigation ("Q" is tentative) */
            sys=type=='N'?SYS_GPS:(type=='L'?SYS_GAL:SYS_QZS);
            if (nav) {
                if (readrnxn(fp,nav,sys)<0) return 'E';
            }
            break;
        case 'G': /* glonass navigation */
            if (nav) {
                if (readrnxg(fp,nav)<0) return 'E';
            }
            break;
        default:
            trace(1,"readrnxfp: unknown rinex type (type=%c)\n",type);
            break;
    }
    return type;
}
/* uncompress and read rinex file --------------------------------------------*/
static char readrnxfile(const char *file, gtime_t ts, gtime_t te, double tint,
                       int rcv, obs_t *obs, nav_t *nav)
{
    FILE *fp;
    int cstat;
    char type,tmpfile[1024],s1[64],s2[64];
    
    time2str(ts,s1,0); time2str(te,s2,0);
    
    trace(3,"readrnxfile: file=%s ts=%s te=%s tint=%.1f rcv=%d\n",file,s1,s2,tint,rcv);
    
    if ((cstat=uncompress(file,tmpfile))<0) {
        trace(1,"readrnxfile: file uncompact error (%s)\n",file);
        return 'E';
    }
    if (!(fp=fopen(cstat?tmpfile:file,"r"))) {
        trace(1,"readrnxfile: file open error (%s)\n",cstat?tmpfile:file);
        return 'E';
    }
    type=readrnxfp(fp,ts,te,tint,rcv,obs,nav);
    fclose(fp);
    
    /* delete temporary file */
    if (cstat) remove(tmpfile);
    return type;
}
/* read rinex files in time range/interval -------------------------------------
* read rinex observation data(obs)/navigation message(nav) files
* args   : char *files[] I      rinex files (wild-card * expanded)
*          int   n       I      number of rinex files (0:input from stdin)
*          gtime_t ts,te I      observation time start/end (ts>=te:all)
*          double tint   I      observation time interval (sec) (0:all)
*          obs_t *obs    IO     observation data    (NULL: no input)
*          nav_t *nav    IO     navigation messages (NULL: no input)
* return : number of epochs (0:no data,<0:error)
*-----------------------------------------------------------------------------*/
extern int readrnxt(char *files[], int n, gtime_t ts, gtime_t te, double tint,
                    obs_t *obs, nav_t *nav)
{
    int i,j,m,rcv=1,nobs=0;
    char type,*efiles[MAXEXFILE]={0},s1[64],s2[64];
    
    time2str(ts,s1,0); time2str(te,s2,0);
    trace(3,"readrnxt: file=%s n=%d ts=%s te=%s tint=%.1f\n",n>0?files[0]:"",n,
          s1,s2,tint);
    
    if (obs) {
        obs->n=obs->nmax=0;
        obs->data=NULL;
    }
    if (nav) {
        nav->n=nav->nmax=nav->ng=nav->ngmax=0;
        nav->eph=NULL; nav->geph=NULL;
    }
    if (n>0) {
        for (i=0;i<MAXEXFILE;i++) {
            if (!(efiles[i]=(char *)malloc(1024))) {
                for (i--;i>=0;i--) free(efiles[i]); return -1;
            }
        }
        for (i=0;i<n;i++) {
            m=expath(files[i],efiles,MAXEXFILE);
            for (j=0,type=' ';j<m;j++) {
                type=readrnxfile(efiles[j],ts,te,tint,rcv,obs,nav);
                if (type=='E') continue;
            }
            if (type=='O') rcv++; /* increment receiver number */
        }
        for (i=0;i<MAXEXFILE;i++) free(efiles[i]);
    }
    else readrnxfp(stdin,ts,te,tint,rcv,obs,nav);
    
    if (obs&&obs->n>0) {
        /* reallocate memory */
        if (!(obs->data=(obsd_t *)realloc(obs->data,sizeof(obsd_t)*obs->n))) {
            obs->n=nav->n=nav->ng=obs->nmax=nav->nmax=0;
            free(nav->eph);
            free(nav->geph);
            return -1;
        }
        obs->nmax=obs->n;
        
        /* if multiple files, sort observation data */
        if (n>1) nobs=sortobs(obs);
    }
    if (nav&&(nav->n>0||nav->ng)) {
        
        /* sort and unique navigation data */
        uniqeph(nav);
        
        /* reallocate memory */
        if (nav->n>0) {
            if (!(nav->eph=(eph_t *)realloc(nav->eph,sizeof(eph_t)*nav->n))) {
                obs->n=nav->n=nav->ng=obs->nmax=nav->nmax=0;
                free(obs->data);
                return -1;
            };
            nav->nmax=nav->n;
        }
        if (nav->ng>0) {
            if (!(nav->geph=(geph_t *)realloc(nav->geph,sizeof(geph_t)*nav->ng))) {
                obs->n=nav->n=nav->ng=obs->nmax=nav->nmax=0;
                free(obs->data);
                return -1;
            };
            nav->ngmax=nav->ng;
        }
    }
    return nobs;
}
/* read rinex files ------------------------------------------------------------
* read rinex observation data(obs)/navigation message(nav) files
* args   : char *files[] I      rinex files (wild-card * expanded)
*          int   n       I      number of rinex files (0:input from stdin)
*          obs_t *obs    O      observation data
*          nav_t *nav    O      navigation messages
* return : number of epochs (0:no data,<0:error)
* note   : obs data are sorted by time, receiver and satellite number
*-----------------------------------------------------------------------------*/
extern int readrnx(char *files[], int n, obs_t *obs, nav_t *nav)
{
    gtime_t t={0};
    
    trace(3,"readrnx: file=%s n=%d\n",n>0?files[0]:"",n);
    
    return readrnxt(files,n,t,t,0.0,obs,nav);
}
/* output rinex obs header -----------------------------------------------------
* output rinex obd file header
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
* return : status (1:ok, 0:output error)
*-----------------------------------------------------------------------------*/
extern int outrnxobsh(FILE *fp, const rnxopt_t *opt)
{
    double eps[6],epe[6],pos[3]={0},del[3]={0};
    int i,j,nt=0;
    char otype[]={'C','L','D','S'},ftype[]={'1','2','5','7','6'};
    char date[32],obstype[32][3]={{""}},*sys;
    
    trace(3,"outrnxobsh: prog=%s runby=%s comment=%s\n",opt->prog,opt->runby,
          opt->comment[0]);
    
    time2str(timeget(),date,0);
    time2epoch(opt->tstart,eps);
    time2epoch(opt->tend,epe);
    if (opt->navsys&~SYS_GPS) sys="M (MIXED)"; else sys="G (GPS)";
    fprintf(fp,"%9.2f%-11s%-20s%-20s%-20s\n",RNXVER,"","OBSERVATION DATA",
            sys,"RINEX VERSION / TYPE");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->prog,opt->runby,date,
            "PGM / RUN BY / DATE");
    for (i=0;i<MAXCOMMENT;i++) {
        if (*opt->comment[i]) fprintf(fp,"%-60.60s%-20s\n",opt->comment[i],"COMMENT");
    }
    fprintf(fp,"%-60.60s%-20s\n",opt->marker,"MARKER NAME");
    fprintf(fp,"%-20.20s%-40.40s%-20s\n",opt->name[0],opt->name[1],"OBSERVER / AGENCY");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->rec[0],opt->rec[1],
            opt->rec[2],"REC # / TYPE / VERS");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->ant[0],opt->ant[1],
            opt->ant[2],"ANT # / TYPE");
    
    for (i=0;i<3;i++) if (fabs(opt->apppos[i])<1E8) pos[i]=opt->apppos[i];
    for (i=0;i<3;i++) if (fabs(opt->antdel[i])<1E8) del[i]=opt->antdel[i];
    fprintf(fp,"%14.4f%14.4f%14.4f%-18s%-20s\n",pos[0],pos[1],pos[2],"",
            "APPROX POSITION XYZ");
    fprintf(fp,"%14.4f%14.4f%14.4f%-18s%-20s\n",del[0],del[1],del[2],"",
            "ANTENNA: DELTA H/E/N");
    fprintf(fp,"%6d%6d%-48s%-20s\n",1,1,"","WAVELENGTH FACT L1/2");
    
    for (i=0;i<NFREQ;i++) for (j=0;j<4;j++) {
        if (!(opt->freqtype&(1<<i))||!(opt->obstype&(1<<j))) continue;
        sprintf(obstype[nt],"%c%c",otype[j],ftype[i]);
        if (!strcmp(obstype[nt],"C2")) strcpy(obstype[nt],"P2");
        if (++nt>=32) break;
    }
    fprintf(fp,"%6d",nt);
    for (i=0;i<nt;i++) {
        if (i>0&&i%9==0) fprintf(fp,"      ");
        fprintf(fp,"%6s",obstype[i]);
        if (i%9==8) fprintf(fp,"%-20s\n","# / TYPES OF OBSERV");
    }
    if (nt==0||i%9>0) {
        fprintf(fp,"%*s%-20s\n",(9-i%9)*6,"","# / TYPES OF OBSERV");
    }
    fprintf(fp,"  %04.0f%6.0f%6.0f%6.0f%6.0f%13.7f     %-12s%-20s\n",eps[0],
            eps[1],eps[2],eps[3],eps[4],eps[5],"GPS","TIME OF FIRST OBS");
    if (opt->tend.time!=0) {
        fprintf(fp,"  %04.0f%6.0f%6.0f%6.0f%6.0f%13.7f     %-12s%-20s\n",epe[0],
                epe[1],epe[2],epe[3],epe[4],epe[5],"","TIME OF LAST OBS");
    }
    return fprintf(fp,"%-60.60s%-20s\n","","END OF HEADER")!=EOF;
}
/* output obs data field -----------------------------------------------------*/
static void outrnxobsf(FILE *fp, double obs, int lli)
{
    if (obs==0.0||obs<=-1E9||obs>=1E9) fprintf(fp,"              ");
    else fprintf(fp,"%14.3f",obs);
    if (lli<=0) fprintf(fp,"  "); else fprintf(fp,"%1.1d ",lli);
}
/* output rinex obs body -------------------------------------------------------
* output rinex obs body
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
*          obsd_t *obs      I   observation data
*          int    n         I   number of observation data
*          int    epflag    I   epoch flag (0:ok,1:power failure,>1:event flag)
* return : status (1:ok, 0:output error)
*-----------------------------------------------------------------------------*/
extern int outrnxobsb(FILE *fp, const rnxopt_t *opt, const obsd_t *obs, int n,
                      int epflag)
{
    double ep[6];
    int i,j,k,nt,ns,prn,sys;
    char sats[MAXOBS][8]={""};
    
    time2epoch(obs[0].time,ep);
    
    for (i=ns=0;i<n&&ns<MAXOBS;i++) {
        if (!(sys=satsys(obs[i].sat,&prn)&opt->navsys)) continue;
        
        switch (sys) {
            case SYS_GPS: sprintf(sats[ns],"G%2d",prn);     break;
            case SYS_GLO: sprintf(sats[ns],"R%2d",prn);     break;
            case SYS_GAL: sprintf(sats[ns],"E%2d",prn);     break;
            case SYS_QZS: sprintf(sats[ns],"Q%2d",prn);     break;
            case SYS_SBS: sprintf(sats[ns],"S%2d",prn-100); break;
        }
        ns++;
    }
    fprintf(fp," %02d %2.0f %2.0f %2.0f %2.0f%11.7f  %d%3d",
            (int)ep[0]%100,ep[1],ep[2],ep[3],ep[4],ep[5],epflag,ns);
    
    for (i=0;i<ns;i++) {
        if (i>0&&i%12==0) fprintf(fp,"\n%32s","");
        fprintf(fp,"%-3s",sats[i]);
    }
    for (i=0;i<n&&ns<MAXOBS;i++) {
        if (!(satsys(obs[i].sat,&prn)&opt->navsys)) continue;
        
        for (j=nt=0;j<NFREQ;j++) for (k=0;k<4;k++) {
            if (!(opt->freqtype&(1<<j))||!(opt->obstype&(1<<k))) continue;
            if (nt%5==0) fprintf(fp,"\n");
            if      (k==0) outrnxobsf(fp,obs[i].P[j],-1);
            else if (k==1) outrnxobsf(fp,obs[i].L[j],obs[i].LLI[j]);
            else if (k==2) outrnxobsf(fp,obs[i].D[j],-1);
            else if (k==3) outrnxobsf(fp,obs[i].SNR[j],-1);
            nt++;
        }
    }
    return fprintf(fp,"\n")!=EOF;
}
/* output nav member by rinex nav format -------------------------------------*/
static void outnavf(FILE *fp, double value)
{
    double e=fabs(value)<1E-99?0.0:floor(log10(fabs(value))+1.0);
    fprintf(fp," %s.%012.0fE%+03.0f",value<0.0?"-":" ",fabs(value)/pow(10.0,e-12.0),e);
}
/* output rinex nav header -----------------------------------------------------
* output rinex nav file header
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
*          double *ion      I   ionospheric parameters (NULL: no output)
*          double *dutc     I   delta-utc parameters   (NULL: no output)
*          int    leaps     I   leap seconds (s) (<0: no output)
* return : status (1:ok, 0:output error)
*-----------------------------------------------------------------------------*/
extern int outrnxnavh(FILE *fp, const rnxopt_t *opt, const double *ion,
                      const double *dutc, int leaps)
{
    int i;
    char date[64];
    
    trace(3,"outrnxnavh: prog=%s runby=%s comment=%s\n",opt->prog,opt->runby,
          opt->comment[0]);
    
    time2str(timeget(),date,0);
    
    fprintf(fp,"%9.2f           %-20s%-20s%-20s\n",RNXVER,"N: GPS NAV DATA","",
            "RINEX VERSION / TYPE");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->prog,opt->runby,date,
            "PGM / RUN BY / DATE");
    
    for (i=0;i<MAXCOMMENT;i++) {
        if (*opt->comment[i]) {
            fprintf(fp,"%-60.60s%-20s\n",opt->comment[i],"COMMENT");
        }
    }
    if (ion) {
        fprintf(fp,"  %12.4E%12.4E%12.4E%12.4E%10s%-20s\n",
                ion[0],ion[1],ion[2],ion[3],"","ION ALPHA");
        fprintf(fp,"  %12.4E%12.4E%12.4E%12.4E%10s%-20s\n",
                ion[4],ion[5],ion[6],ion[7],"","ION BETA");
    }
    if (dutc) {
        fprintf(fp,"   ");
        outnavf(fp,dutc[0]);
        outnavf(fp,dutc[1]);
        fprintf(fp,"%9.0f%9.0f %-20s\n",dutc[2],dutc[3],"DELTA-UTC: A0,A1,T,W");
    }
    if (leaps>=0) {
        fprintf(fp,"%6d%54s%-20s\n",leaps,"","LEAP SECONDS");
    }
    return fprintf(fp,"%60s%-20s\n","","END OF HEADER")!=EOF;
}
/* output rinex nav body -------------------------------------------------------
* output rinex nav file body record
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
*          eph_t  *eph      I   ephemeris record
* return : status (1:ok, 0:output error)
*-----------------------------------------------------------------------------*/
extern int outrnxnavb(FILE *fp, const rnxopt_t *opt, const eph_t *eph)
{
    double t[6],ttr;
    int week,w;
    
    (void)time2gpst(eph->toe,&week);
    time2epoch(eph->toc,t);
    fprintf(fp,"%2d %02d %2.0f %2.0f %2.0f %2.0f %4.1f",eph->sat,(int)t[0]%100,
            t[1],t[2],t[3],t[4],t[5]);
    
    outnavf(fp,eph->f0    );
    outnavf(fp,eph->f1    );
    outnavf(fp,eph->f2    );  fprintf(fp,"\n   ");
    
    outnavf(fp,eph->iode  );
    outnavf(fp,eph->crs   );
    outnavf(fp,eph->deln  );
    outnavf(fp,eph->M0    );  fprintf(fp,"\n   ");
    
    outnavf(fp,eph->cuc   );
    outnavf(fp,eph->e     );
    outnavf(fp,eph->cus   );
    outnavf(fp,sqrt(eph->A)); fprintf(fp,"\n   ");
    
    outnavf(fp,eph->toes  );
    outnavf(fp,eph->cic   );
    outnavf(fp,eph->OMG0  );
    outnavf(fp,eph->cis   );  fprintf(fp,"\n   ");
    
    outnavf(fp,eph->i0    );
    outnavf(fp,eph->crc   );
    outnavf(fp,eph->omg   );
    outnavf(fp,eph->OMGd  );  fprintf(fp,"\n   ");
    
    outnavf(fp,eph->idot  );
    outnavf(fp,eph->l2code);
    outnavf(fp,week       );
    outnavf(fp,eph->l2pflg);  fprintf(fp,"\n   ");
    
    outnavf(fp,eph->sva   );
    outnavf(fp,eph->svh   );
    outnavf(fp,eph->tgd   );
    outnavf(fp,eph->iodc  );  fprintf(fp,"\n   ");
    
    ttr=time2gpst(eph->ttr,&w);
    outnavf(fp,ttr+(w-week)*604800.0);
    outnavf(fp,eph->fit );
    
    return fprintf(fp,"\n")!=EOF;
}
/* output rinex gnav header ----------------------------------------------------
* output rinex gnav (glonass navigation) file header
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
* return : status (1:ok, 0:output error)
*-----------------------------------------------------------------------------*/
extern int outrnxgnavh(FILE *fp, const rnxopt_t *opt)
{
    int i;
    char date[64];
    
    trace(3,"outrnxgnavh: prog=%s runby=%s comment=%s\n",opt->prog,opt->runby,
          opt->comment[0]);
    
    time2str(timeget(),date,0);
    
    fprintf(fp,"%9.2f           %-20s%-20s%-20s\n",RNXVER,"GLONASS NAV DATA","",
            "RINEX VERSION / TYPE");
    fprintf(fp,"%-20.20s%-20.20s%-20.20s%-20s\n",opt->prog,opt->runby,date,
            "PGM / RUN BY / DATE");
    
    for (i=0;i<MAXCOMMENT;i++) {
        if (*opt->comment[i]) {
            fprintf(fp,"%-60.60s%-20s\n",opt->comment[i],"COMMENT");
        }
    }
    return fprintf(fp,"%60s%-20s\n","","END OF HEADER")!=EOF;
}
/* output rinex gnav body ------------------------------------------------------
* output rinex gnav (glonass navigation) file body record
* args   : FILE   *fp       I   output file pointer
*          rnxopt_t *opt    I   rinex options
*          geph_t  *geph    I   glonass ephemeris record
* return : status (1:ok, 0:output error)
*-----------------------------------------------------------------------------*/
extern int outrnxgnavb(FILE *fp, const rnxopt_t *opt, const geph_t *geph)
{
    gtime_t toe;
    double ep[6],tof;
    int prn;
    
    if (satsys(geph->sat,&prn)!=SYS_GLO) return 0;
    
    tof=fmod(time2gpst(gpst2utc(geph->tof),NULL),86400.0);
    toe=gpst2utc(geph->toe); /* gpst -> utc */
    time2epoch(toe,ep);
    fprintf(fp,"%2d %02d %2.0f %2.0f %2.0f %2.0f %4.1f",prn,(int)ep[0]%100,
            ep[1],ep[2],ep[3],ep[4],ep[5]);
    
    outnavf(fp,-geph->taun );
    outnavf(fp,geph->gamn  );
    outnavf(fp,tof         ); fprintf(fp,"\n   ");
    
    outnavf(fp,geph->pos[0]/1E3);
    outnavf(fp,geph->vel[0]/1E3);
    outnavf(fp,geph->acc[0]/1E3);
    outnavf(fp,geph->svh   ); fprintf(fp,"\n   ");
    
    outnavf(fp,geph->pos[1]/1E3);
    outnavf(fp,geph->vel[1]/1E3);
    outnavf(fp,geph->acc[1]/1E3);
    outnavf(fp,geph->frq   ); fprintf(fp,"\n   ");
    
    outnavf(fp,geph->pos[2]/1E3);
    outnavf(fp,geph->vel[2]/1E3);
    outnavf(fp,geph->acc[2]/1E3);
    outnavf(fp,geph->age   );
    
    return fprintf(fp,"\n")!=EOF;
}
/* uncompress file -------------------------------------------------------------
* uncompress (uncompress/unzip/uncompact hatanaka-compression) file
* args   : char   *file     I   input file
*          char   *uncfile  O   uncompressed file
* return : status (-1:error,0:not compressed file,1:uncompress completed)
* note   : creates uncompressed file in tempolary directory
*          gzip and crx2rnx command have to be installed in commands path
*-----------------------------------------------------------------------------*/
extern int uncompress(const char *file, char *uncfile)
{
    int stat=0;
    char *p,cmd[1024]="",tmpfile[1024]="";
    
    trace(3,"uncompress: file=%s\n",file);
    
    strcpy(tmpfile,file);
    if (!(p=strrchr(tmpfile,'.'))) return 0;
    
    /* uncompress by gzip */
    if (!strcmp(p,".z"  )||!strcmp(p,".Z"  )||
        !strcmp(p,".gz" )||!strcmp(p,".GZ" )||
        !strcmp(p,".zip")||!strcmp(p,".ZIP")) {
        
        strcpy(uncfile,tmpfile); uncfile[p-tmpfile]='\0';
        sprintf(cmd,"gzip -f -d -c \"%s\" > \"%s\"",tmpfile,uncfile);
        
        if (execcmd(cmd)) {
            remove(uncfile);
            return -1;
        }
        strcpy(tmpfile,uncfile);
        stat=1;
    }
    /* extract hatanaka-compressed file by cnx2rnx */
    if ((p=strrchr(tmpfile,'.'))&&strlen(p)>3&&(*(p+3)=='d'||*(p+3)=='D')) {
        
        strcpy(uncfile,tmpfile);
        uncfile[p-tmpfile+3]=*(p+3)=='D'?'O':'o';
        sprintf(cmd,"crx2rnx < \"%s\" > \"%s\"",tmpfile,uncfile);
        
        if (execcmd(cmd)) {
            remove(uncfile);
            if (stat) remove(tmpfile);
            return -1;
        }
        if (stat) remove(tmpfile);
        stat=1;
    }
    trace(3,"uncompress: stat=%d\n",stat);
    return stat;
}
